@@ -99,39 +99,6 @@ static const uint16_t char_table[256] = {
   0x1801, 0x0f0a, 0x103f, 0x203f, 0x2001, 0x0f0b, 0x1040, 0x2040
 };
 
-
-static inline const char*
-varint_parse32(const char *p, const char *l, uint32_t *OUTPUT)
-{
-	const uint8_t *ptr = (const uint8_t*)p;
-	const uint8_t *limit = (const uint8_t*)l;
-	uint32_t b, result;
-	if (ptr >= limit) return NULL;
-	b = *(ptr++);
-	result = b & 127;
-	if (b < 128) goto done;
-	if (ptr >= limit) return NULL;
-	b = *(ptr++);
-	result |= (b & 127) << 7;
-	if (b < 128) goto done;
-	if (ptr >= limit) return NULL;
-	b = *(ptr++);
-	result |= (b & 127) << 14;
-	if (b < 128) goto done;
-	if (ptr >= limit) return NULL;
-	b = *(ptr++);
-	result |= (b & 127) << 21;
-	if (b < 128) goto done;
-	if (ptr >= limit) return NULL;
-	b = *(ptr++);
-	result |= (b & 127) << 28;
-	if (b < 16) goto done;
-	return NULL; /* Value is too long to be a varint32 */
-done:
-	*OUTPUT = result;
-	return (const char*)ptr;
-}
-
 /*
  * Copy "len" bytes from "src" to "op", one byte at a time.  Used for
  * handling COPY operations where the input and output regions may
@@ -209,25 +176,6 @@ struct SnappyArrayWriter {
 	char* op_limit;
 };
 
-static inline void
-SAW__init(struct SnappyArrayWriter *this, char *dst)
-{
-	this->base = dst;
-	this->op = dst;
-}
-
-static inline void
-SAW__SetExpectedLength(struct SnappyArrayWriter *this, size_t len)
-{
-	this->op_limit = this->op + len;
-}
-
-static inline int
-SAW__CheckLength(const struct SnappyArrayWriter *this)
-{
-	return this->op == this->op_limit;
-}
-
 static inline int
 SAW__Append(struct SnappyArrayWriter *this,
 	    const char* ip, uint32_t len, int allow_fast_path)
@@ -240,7 +188,7 @@ SAW__Append(struct SnappyArrayWriter *this,
 		UNALIGNED_STORE64(op + 8, UNALIGNED_LOAD64(ip + 8));
 	} else {
 		if (space_left < len)
-			return FALSE;
+			return SNAPPY_E_OUTPUT_OVERRUN;
 		memcpy(op, ip, len);
 	}
 	this->op = op + len;
@@ -255,19 +203,17 @@ SAW__AppendFromSelf(struct SnappyArrayWriter *this,
 	const int space_left = this->op_limit - op;
 	/* -1u catches offset==0 */
 	if (op - this->base <= offset - 1u)
-		return FALSE;
+		return SNAPPY_E_DATA_MALFORMED;
 	/* Fast path, used for the majority (70-80%) of dynamic invocations. */
 	if (len <= 16 && offset >= 8 && space_left >= 16) {
 		UNALIGNED_STORE64(op, UNALIGNED_LOAD64(op - offset));
 		UNALIGNED_STORE64(op + 8, UNALIGNED_LOAD64(op - offset + 8));
+	} else if (space_left >= len + kMaxIncrementCopyOverflow) {
+		IncrementalCopyFastPath(op - offset, op, len);
 	} else {
-		if (space_left >= len + kMaxIncrementCopyOverflow) {
-			IncrementalCopyFastPath(op - offset, op, len);
-		} else {
-			if (space_left < len)
-				return FALSE;
-			IncrementalCopy(op - offset, op, len);
-		}
+		if (space_left < len)
+			return SNAPPY_E_OUTPUT_OVERRUN;
+		IncrementalCopy(op - offset, op, len);
 	}
 	this->op = op + len;
 	return TRUE;
@@ -277,16 +223,16 @@ SAW__AppendFromSelf(struct SnappyArrayWriter *this,
 /* Helper class for decompression */
 struct SnappyDecompressor {
 	const char	*src;
-	size_t		src_bytes_left;
-	const char		*ip;		/* Points to next buffered byte */
-	const char		*ip_limit;	/* Points just past buffered bytes */
-	uint32_t		peeked;		/* Bytes peeked from reader (need to skip) */
-	int			eof;		/* Hit end of input without an error? */
-	char			scratch[5];	/* Temporary buffer for PeekFast() boundaries */
+	uint32_t	src_bytes_left;
+	const char	*ip;		/* Points to next buffered byte */
+	const char	*ip_limit;	/* Points just past buffered bytes */
+	uint32_t	peeked;		/* Bytes peeked from reader (need to skip) */
+	int		eof;		/* Hit end of input without an error? */
+	char		scratch[5];	/* Temporary buffer for PeekFast() boundaries */
 };
 
 static inline void
-SD__init(struct SnappyDecompressor *this, const char *source, size_t src_len)
+SD__init(struct SnappyDecompressor *this, const char *source, uint32_t src_len)
 {
 	this->src = source;
 	this->src_bytes_left = src_len;
@@ -309,12 +255,13 @@ static inline uint32_t MIN_UINT32(uint32_t a, uint32_t b)
  * in [ip_..ip_limit_-1].  Also ensures that [ip,ip+4] is readable even
  * if (ip_limit_ - ip_ < 5).
  *
- * Returns TRUE on success, FALSE on error or end of input.
+ * Returns TRUE on success.
  */
 static int
 SD__RefillTag(struct SnappyDecompressor *this)
 {
 	const char* ip = this->ip;
+	uint32_t needed, nbuf, to_add;
 	if (ip == this->ip_limit) {
 		/* Fetch a new fragment from the reader */
 		/* All peeked bytes are used up */
@@ -324,20 +271,19 @@ SD__RefillTag(struct SnappyDecompressor *this)
 		this->peeked = this->src_bytes_left;
 		if (this->src_bytes_left == 0) {
 			this->eof = TRUE;
-			return FALSE;
+			return SNAPPY_E_OK;
 		}
 		this->ip_limit = ip + this->src_bytes_left;
 	}
 
 	/* Read the tag character */
 	DCHECK_LT(ip, this->ip_limit);
-	const uint8_t c = *(const uint8_t*)ip;
-	const uint32_t entry = char_table[c];
-	const uint32_t needed = (entry >> 11) + 1; /* +1 byte for 'c' */
+	/* +1 byte for current byte at ip */
+	needed = (char_table[*(const uint8_t*)ip] >> 11) + 1;
 	DCHECK_LE(needed, sizeof(this->scratch));
 
 	/* Read more bytes from reader if needed */
-	uint32_t nbuf = this->ip_limit - ip;
+	nbuf = this->ip_limit - ip;
 	if (nbuf < needed) {
 		/* Stitch together bytes from ip and reader to form the word
 		   contents. We store the needed bytes in "scratch_". They
@@ -350,8 +296,8 @@ SD__RefillTag(struct SnappyDecompressor *this)
 		this->peeked = 0;
 		while (nbuf < needed) {
 			if (this->src_bytes_left == 0)
-				return FALSE;
-			uint32_t to_add = MIN_UINT32(needed - nbuf, this->src_bytes_left);
+				return SNAPPY_E_OUTPUT_OVERRUN;
+			to_add = MIN_UINT32(needed - nbuf, this->src_bytes_left);
 			memcpy(this->scratch + nbuf, this->src, to_add);
 			nbuf += to_add;
 			this->src += to_add;
@@ -377,31 +323,25 @@ SD__RefillTag(struct SnappyDecompressor *this)
 	return TRUE;
 }
 
-/* Returns TRUE iff we have hit the end of the input without an error. */
-static inline int
-SD__eof(const struct SnappyDecompressor *this)
-{
-	return this->eof;
-}
-
 /*
  * Read the uncompressed length stored at the start of the compressed data.
- * On succcess, stores the length in *result and returns TRUE.
- * On failure, returns FALSE.
+ * On succcess, stores the length in *result and returns SNAPPY_E_OK.
+ * On failure, returns SNAPPY_E_HEADER_BAD.
  */
-static inline int
+static noinline int
 SD__ReadUncompressedLength(struct SnappyDecompressor *this, uint32_t *result)
 {
+	uint32_t shift = 0;
+	uint8_t c;
 	DCHECK(this->ip == NULL); /* Must not have read anything yet */
 	/* Length is encoded in 1..5 bytes */
 	*result = 0;
-	uint32_t shift = 0;
 	for(;;) {
 		if (shift >= 32)
-			return FALSE;
+			return SNAPPY_E_HEADER_BAD;
 		if (this->src_bytes_left == 0)
-			return FALSE;
-		const uint8_t c = *(const uint8_t*)this->src;
+			return SNAPPY_E_HEADER_BAD;
+		c = *(const uint8_t*)this->src;
 		this->src += 1;
 		this->src_bytes_left -= 1;
 		*result |= (uint32_t)(c & 0x7f) << shift;
@@ -409,37 +349,41 @@ SD__ReadUncompressedLength(struct SnappyDecompressor *this, uint32_t *result)
 			break;
 		shift += 7;
 	}
-	return TRUE;
+	return SNAPPY_E_OK;
 }
 
 /*
  * Process the next item found in the input.
- * Returns TRUE if successful, FALSE on error or end of input.
+ * Returns TRUE if more data is pending.
  */
 static inline int
 SD__Step(struct SnappyDecompressor *this, struct SnappyArrayWriter *writer)
 {
+	uint8_t c;
+	uint32_t entry, trailer, length, literal_length, avail;
+	int ret, allow_fast_path;
 	const char* ip = this->ip;
 	if (this->ip_limit - ip < 5) {
-		if (!SD__RefillTag(this))
-			return FALSE;
+		if ((ret = SD__RefillTag(this)) != TRUE)
+			return ret;
 		ip = this->ip;
 	}
 
-	const uint8_t c = *(const uint8_t*)(ip++);
-	const uint32_t entry = char_table[c];
-	const uint32_t trailer = le32_to_cpu(UNALIGNED_LOAD32(ip)) &
-				 wordmask[entry >> 11];
+	c = *(const uint8_t*)(ip++);
+	entry = char_table[c];
+	trailer = le32_to_cpu(UNALIGNED_LOAD32(ip)) &
+				wordmask[entry >> 11];
 	ip += entry >> 11;
-	const uint32_t length = entry & 0xff;
+	length = entry & 0xff;
 
 	if ((c & 0x3) == LITERAL) {
-		uint32_t literal_length = length + trailer;
-		uint32_t avail = this->ip_limit - ip;
+		literal_length = length + trailer;
+		avail = this->ip_limit - ip;
 		while (avail < literal_length) {
-			int allow_fast_path = (avail >= 16);
-			if (!SAW__Append(writer, ip, avail, allow_fast_path))
-				return FALSE;
+			allow_fast_path = (avail >= 16);
+			ret = SAW__Append(writer, ip, avail, allow_fast_path);
+			if (ret != TRUE)
+				return ret;
 			literal_length -= avail;
 			this->src += this->peeked;
 			this->src_bytes_left -= this->peeked;
@@ -447,60 +391,59 @@ SD__Step(struct SnappyDecompressor *this, struct SnappyArrayWriter *writer)
 			avail = this->src_bytes_left;
 			this->peeked = avail;
 			if (avail == 0)
-				return FALSE; /* Premature end of input */
+				return SNAPPY_E_INPUT_NOT_CONSUMED;
 			this->ip_limit = ip + avail;
 		}
 		this->ip = ip + literal_length;
-		int allow_fast_path = (avail >= 16);
+		allow_fast_path = (avail >= 16);
 		return SAW__Append(writer, ip, literal_length, allow_fast_path);
 	} else {
 		this->ip = ip;
 		/* copy_offset/256 is encoded in bits 8..10.  By just fetching
 		   those bits, we get copy_offset (since the bit-field starts at
 		   bit 8). */
-		const uint32_t copy_offset = entry & 0x700;
-		return SAW__AppendFromSelf(writer, copy_offset + trailer, length);
+		return SAW__AppendFromSelf(writer, (entry & 0x700) + trailer, length);
 	}
 }
 
 
 int
-snappy_get_uncompressed_length(const char *start, size_t n, size_t *result)
+snappy_get_uncompressed_length(const char *start, uint32_t n, uint32_t *result)
 {
-	uint32_t v = 0;
-	const char *limit = start + n;
-	if (varint_parse32(start, limit, &v) != NULL) {
-		*result = v;
-		return TRUE;
-	} else {
-		return FALSE;
-	}
+	struct SnappyDecompressor decomp;
+	SD__init(&decomp, start, n);
+	return SD__ReadUncompressedLength(&decomp, result);
 }
 #if defined(__KERNEL__) && !defined(STATIC)
 EXPORT_SYMBOL(snappy_get_uncompressed_length);
 #endif
 
 int
-snappy_decompress(const char *src, size_t src_len, char *dst, size_t dst_len)
+snappy_decompress(const char *src, uint32_t src_len, char *dst, uint32_t dst_len)
 {
 	struct SnappyArrayWriter writer;
-	struct SnappyDecompressor decompressor;
-	SAW__init(&writer, dst);
+	struct SnappyDecompressor decomp;
+	int ret;
+	uint32_t olen = 0;
+	writer.base = writer.op = dst;
+	SD__init(&decomp, src, src_len);
 	/* Read the uncompressed length from the front of the compressed input */
-	SD__init(&decompressor, src, src_len);
-	uint32_t uncompressed_len = 0;
-	if (!SD__ReadUncompressedLength(&decompressor, &uncompressed_len))
-		goto error;
+	ret = SD__ReadUncompressedLength(&decomp, &olen);
+	if (unlikely(ret != SNAPPY_E_OK))
+		return ret;
 	/* Protect against possible DoS attack */
-	if ((size_t)uncompressed_len > dst_len)
-		goto error;
-	SAW__SetExpectedLength(&writer, uncompressed_len);
+	if (unlikely(olen > dst_len))
+		return SNAPPY_E_OUTPUT_INSUF;
+	writer.op_limit = writer.op + olen;
 	/* Process the entire input */
-	while (SD__Step(&decompressor, &writer)) { }
-	int status = (SD__eof(&decompressor) && SAW__CheckLength(&writer));
-	if (status) return TRUE; else return FALSE;
-error:
-	return FALSE;
+	while ((ret = SD__Step(&decomp, &writer)) == TRUE) { }
+	if (ret != SNAPPY_E_OK)
+		return ret;
+	if (decomp.eof != TRUE)
+		return SNAPPY_E_INPUT_NOT_CONSUMED;
+	if (writer.op != writer.op_limit)
+		return SNAPPY_E_UNEXPECTED_OUTPUT_LEN;
+	return SNAPPY_E_OK;
 }
 #if defined(__KERNEL__) && !defined(STATIC)
 EXPORT_SYMBOL(snappy_decompress);
@@ -508,80 +451,3 @@ EXPORT_SYMBOL(snappy_decompress);
 MODULE_LICENSE("BSD");
 MODULE_DESCRIPTION("Snappy Decompressor");
 #endif
-
-#ifdef TEST
-#define MAX_INPUT_SIZE 10 * 1024 * 1024
-#include <stdio.h>
-int main(int argc, char *argv[])
-{
-	FILE *input_file, *output_file;
-	if (argc < 3) {
-		fprintf(stderr, "Usage: first argument is input file, "
-				"second argument is output file.\n"
-				"Use - for stdin/stdout.\n");
-		return 1;
-	}
-	if (strcmp("-", argv[1]) == 0)
-		input_file = stdin;
-	else
-		input_file = fopen(argv[1], "rb");
-	if (strcmp("-", argv[2]) == 0)
-		output_file = stdout;
-	else
-		output_file = fopen(argv[2], "wb");
-	
-	char *input_bufer = (char *)malloc(MAX_INPUT_SIZE);
-	if (!input_bufer)
-	{
-		fprintf(stderr, "malloc failed to allocate %d.\n", MAX_INPUT_SIZE);
-		fclose(input_file);
-		fclose(output_file);
-		return 2;
-	}
-	size_t input_len = fread(input_bufer, 1, MAX_INPUT_SIZE, input_file);
-	if (!feof(input_file))
-	{
-		fprintf(stderr, "input was longer than %d, aborting.\n", MAX_INPUT_SIZE);
-		free(input_bufer);
-		fclose(input_file);
-		fclose(output_file);
-		return 3;
-	}
-	fclose(input_file);
-	
-	size_t uncompressed_len;
-	if (snappy_get_uncompressed_length(input_bufer, input_len, &uncompressed_len) == FALSE)
-	{
-		fprintf(stderr, "snappy_get_uncompressed_length failed.\n");
-		free(input_bufer);
-		fclose(output_file);
-		return 4;
-	}
-	
-	char* output_buffer = (char *)malloc(uncompressed_len);
-	if (!output_buffer)
-	{
-		fprintf(stderr, "malloc failed to allocate %d.\n", (int)uncompressed_len);
-		free(input_bufer);
-		fclose(output_file);
-		return 2;
-	}
-	
-	int status = snappy_decompress(input_bufer, input_len, output_buffer, uncompressed_len);
-	free(input_bufer);
-	if (status == FALSE)
-	{
-		fprintf(stderr, "snappy_decompress failed.\n");
-		free(output_buffer);
-		fclose(output_file);
-		return 5;
-	}
-	
-	fwrite(output_buffer, 1, uncompressed_len, output_file);
-	fclose(output_file);
-	
-	free(output_buffer);
-	
-	return 0;
-}
-#endif
